#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Contains the metadata for all the commands.

The metadata represents the following:
    - table_params: Parameters required for the table output formatting.
    - request_body_yaml_path: File containing the yaml file examples for the command.
"""

from __future__ import annotations

from typing import Any, Dict, List


def tag_parser(data: Any) -> str:
    tag_array = [tag["name"] for tag in data]
    return ", ".join(tag_array)


command_metadata: Dict[str, Any] = {
    "images>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "name": {
                    "label": "Image Name",
                },
                "type": {
                    "label": "Type",
                },
                "scope": {
                    "label": "Scope",
                },
                "latestVersion.id": {
                    "label": "Latest Version ID",
                },
                "latestVersion.version": {
                    "label": "Latest Version",
                },
                "latestVersion.status": {
                    "label": "Latest Version Status",
                },
                "latestVersion.lastBuildStatus": {
                    "label": "Latest Version Build Status",
                },
                "latestVersion.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Images",
            "data_key": "images",
            "subheader_key": "imageCount",
        },
    },
    "images>list-versions": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "version": {
                    "label": "Version",
                },
                "description": {
                    "label": "Description",
                },
                "status": {
                    "label": "Status",
                },
                "lastBuildStatus": {
                    "label": "Last Build Status",
                },
                "buildDetails": {
                    "label": "Build Details",
                },
                "tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Image Versions",
            "data_key": "versions",
            "subheader_key": "versionCount",
        },
    },
    "images>list-builds": {
        "table_params": {
            "output_keys": {
                "buildId": {
                    "label": "Build ID",
                },
                "versionId": {
                    "label": "Version ID",
                },
                "version": {
                    "label": "Version",
                },
                "status": {
                    "label": "Build Status",
                },
                "startedAt": {
                    "label": "Started At (UTC)",
                },
                "finishedAt": {
                    "label": "Finished At (UTC)",
                },
            },
            "title": "Image Builds",
            "data_key": "builds",
            "subheader_key": "buildCount",
        },
    },
    "workflows>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "name": {
                    "label": "Workflow Name",
                },
                "status": {
                    "label": "Status",
                },
                "triggers": {
                    "label": "Triggers",
                },
                "lastExecution.status": {
                    "label": "Last Execution Status",
                },
                "lastExecution.executedAt": {
                    "label": "Last Execution Time (UTC)",
                },
                "tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Workflows",
            "data_key": "workflows",
            "subheader_key": "workflowsCount",
        },
    },
    "workflows>list-executions": {
        "table_params": {
            "output_keys": {
                "executionId": {
                    "label": "Execution ID",
                },
                "status": {
                    "label": "Status",
                },
                "duration": {
                    "label": "Run Duration",
                },
                "executedAt": {
                    "label": "Execution Time (UTC)",
                },
            },
            "title": "Workflows Executions",
            "data_key": "executions",
            "subheader_key": "executionsCount",
        },
    },
    "workflows>get-execution-details": {
        "table_params": {
            "output_keys": {
                "name": {
                    "label": "Step Name",
                },
                "startedAt": {
                    "label": "Started At",
                },
                "finishedAt": {
                    "label": "Finished At",
                },
                "status": {
                    "label": "Status",
                },
                "stepType": {
                    "label": "Step Type",
                },
                "image": {
                    "label": "Image",
                },
                "output": {
                    "label": "Output",
                },
            },
            "title": "Execution Details",
            "data_key": "steps",
            "subheader_title": "Execution Status",
            "subheader_key": "status",
        },
    },
    "webapps>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "name": {
                    "label": "Name",
                },
                "status": {
                    "label": "Status",
                },
                "updatedBy": {
                    "label": "Updated By",
                },
                "updatedAt": {
                    "label": "Updated At (UTC)",
                },
                "tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Webapps",
            "data_key": "webapps",
            "subheader_key": "webappsCount",
        },
    },
    "services>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "name": {
                    "label": "Name",
                },
                "serviceType": {
                    "label": "Service Type",
                },
                "status": {
                    "label": "Status",
                },
                "updatedBy": {
                    "label": "Updated By",
                },
                "updatedAt": {
                    "label": "Updated At (UTC)",
                },
                "tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Services",
            "data_key": "services",
            "subheader_key": "servicesCount",
        },
    },
    "artifacts>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "description": {
                    "label": "Description",
                },
                "name": {
                    "label": "Name",
                },
                "createdAt": {
                    "label": "Created At (UTC)",
                },
                "createdBy": {
                    "label": "Created By",
                },
            },
            "title": "Artifacts",
            "data_key": "artifacts",
            "subheader_key": "artifactCount",
        },
    },
    "specs>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "kind": {
                    "label": "Kind",
                },
                "metadata.name": {
                    "label": "Name",
                },
                "metadata.status": {
                    "label": "Status",
                },
                "latestRelease.version": {
                    "label": "Latest Release",
                },
                "metadata.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Specs",
            "data_key": "specs",
            "subheader_key": "specCount",
        },
    },
    "specs>list-release-deployments": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "kind": {
                    "label": "Kind",
                },
                "metadata.name": {
                    "label": "Name",
                },
                "metadata.status": {
                    "label": "Status",
                },
                "latestRevision.revision": {
                    "label": "Latest Revision",
                },
                "latestRevision.status": {
                    "label": "Latest Revision Status",
                },
                "metadata.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Deployments for a Spec Release",
            "data_key": "deployments",
            "subheader_key": "deploymentCount",
        },
    },
    "deployments>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "kind": {
                    "label": "Kind",
                },
                "metadata.title": {
                    "label": "Title",
                },
                "metadata.status": {
                    "label": "Status",
                },
                "latestRevision.revision": {
                    "label": "Latest Revision",
                },
                "latestRevision.status": {
                    "label": "Latest Revision Status",
                },
                "metadata.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Deployments",
            "data_key": "deployments",
            "subheader_key": "deploymentCount",
        },
    },
    "apps>specs>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "kind": {
                    "label": "Kind",
                },
                "metadata.name": {
                    "label": "Name",
                },
                "metadata.status": {
                    "label": "Status",
                },
                "latestRelease.version": {
                    "label": "Latest Release",
                },
                "metadata.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Specs",
            "data_key": "specs",
            "subheader_key": "specCount",
        },
    },
    "apps>specs>list-releases": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "version": {
                    "label": "Version",
                },
                "notes": {
                    "label": "Notes",
                },
                "createdAt": {
                    "label": "Created At (UTC)",
                },
                "createdBy": {
                    "label": "Created By",
                },
            },
            "title": "Spec Releases",
            "data_key": "releases",
            "subheader_key": "releaseCount",
        },
    },
    "apps>deployments>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "kind": {
                    "label": "Kind",
                },
                "metadata.title": {
                    "label": "Title",
                },
                "metadata.status": {
                    "label": "Status",
                },
                "latestRevision.revision": {
                    "label": "Latest Revision",
                },
                "latestRevision.status": {
                    "label": "Latest Revision Status",
                },
                "metadata.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "App Deployments",
            "data_key": "deployments",
            "subheader_key": "deploymentCount",
        },
    },
    "apps>deployments>list-revisions": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "notes": {
                    "label": "Notes",
                },
                "revision": {
                    "label": "Revision",
                },
                "status": {
                    "label": "Status",
                },
            },
            "title": "App Deployment Revisions",
            "data_key": "revisions",
            "subheader_key": "revisionCount",
        },
    },
    "blocks>specs>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "kind": {
                    "label": "Kind",
                },
                "metadata.name": {
                    "label": "Name",
                },
                "metadata.status": {
                    "label": "Status",
                },
                "latestRelease.version": {
                    "label": "Latest Release",
                },
                "metadata.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Block Specs",
            "data_key": "specs",
            "subheader_key": "specCount",
        },
    },
    "blocks>specs>list-releases": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "version": {
                    "label": "Version",
                },
                "notes": {
                    "label": "Notes",
                },
                "createdAt": {
                    "label": "Created At (UTC)",
                },
                "createdBy": {
                    "label": "Created By",
                },
            },
            "title": "Block Spec Releases",
            "data_key": "releases",
            "subheader_key": "releaseCount",
        },
    },
    "blocks>deployments>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "kind": {
                    "label": "Kind",
                },
                "metadata.title": {
                    "label": "Title",
                },
                "metadata.status": {
                    "label": "Status",
                },
                "latestRevision.revision": {
                    "label": "Latest Revision",
                },
                "latestRevision.status": {
                    "label": "Latest Revision Status",
                },
                "metadata.tags": {
                    "label": "Tags",
                    "parser": tag_parser,
                },
            },
            "title": "Block Deployments",
            "data_key": "deployments",
            "subheader_key": "deploymentCount",
        },
    },
    "blocks>deployments>list-revisions": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "notes": {
                    "label": "Notes",
                },
                "revision": {
                    "label": "Revision",
                },
                "status": {
                    "label": "Status",
                },
            },
            "title": "Block Deployment Revisions",
            "data_key": "revisions",
            "subheader_key": "revisionCount",
        },
    },
    "deployments>execute-resources": {
        "table_params": {
            "output_keys": {
                "blockSpecId": {
                    "label": "Block Spec ID",
                },
                "version": {
                    "label": "Version",
                },
                "executionId": {
                    "label": "Execution ID",
                },
                "status": {
                    "label": "Execution Status",
                },
            },
            "title": "Trigger Resources",
            "data_key": "executeResponse",
        },
    },
    "tenants>list-instance-options": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "name": {
                    "label": "Instance Name",
                },
                "cpu": {
                    "label": "CPU",
                },
                "memory": {
                    "label": "Memory",
                },
                "gpu": {
                    "label": "GPU",
                },
                "gpuMemory": {
                    "label": "GPU Memory",
                },
                "provider": {
                    "label": "Provider",
                },
            },
            "title": "Instance Options",
            "data_key": "data",
        },
    },
    "alerts>emails>list": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "subject": {
                    "label": "Subject",
                },
                "status": {
                    "label": "Status",
                },
                "templateName": {
                    "label": "Template Name",
                },
                "createdAt": {
                    "label": "Created At",
                },
                "createdBy": {
                    "label": "Created By",
                },
            },
            "title": "Emails",
            "data_key": "emails",
            "subheader_key": "emailCount",
        },
    },
    "alerts>emails>list-templates": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "name": {
                    "label": "Name",
                },
                "scope": {
                    "label": "Status",
                },
                "createdAt": {
                    "label": "Created At",
                },
                "createdBy": {
                    "label": "Created By",
                },
            },
            "title": "Templates",
            "data_key": "templates",
            "subheader_key": "templateCount",
        },
    },
    "metrics>list": {
        "table_params": {
            "output_keys": {
                "name": {
                    "label": "Name",
                },
                "type": {
                    "label": "Type",
                },
                "description": {
                    "label": "Description",
                },
                "collectionId": {
                    "label": "Collection ID",
                },
                "publicationId": {
                    "label": "Publication ID",
                },
                "metadata": {
                    "label": "Metadata",
                },
            },
            "title": "Metrics",
            "data_key": "data",
            "subheader_key": "totalCount",
        },
    },
    "metrics>list>measure": {
        "table_params": {
            "output_keys": {
                "name": {
                    "label": "Name",
                },
                "type": {
                    "label": "Type",
                },
                "aggType": {
                    "label": "Aggregation Type",
                },
                "cube": {
                    "label": "Cube",
                },
                "description": {
                    "label": "Description",
                },
                "collectionId": {
                    "label": "Collection ID",
                },
                "publicationId": {
                    "label": "Publication ID",
                },
                "metadata": {
                    "label": "Metadata",
                },
            },
            "title": "Measures",
            "data_key": "data",
            "subheader_key": "totalCount",
        },
    },
    "metrics>list>dimension": {
        "table_params": {
            "output_keys": {
                "name": {
                    "label": "Name",
                },
                "type": {
                    "label": "Type",
                },
                "primaryKey": {
                    "label": "Primary Key",
                },
                "cube": {
                    "label": "Cube",
                },
                "description": {
                    "label": "Description",
                },
                "collectionId": {
                    "label": "Collection ID",
                },
                "publicationId": {
                    "label": "Publication ID",
                },
                "metadata": {
                    "label": "Metadata",
                },
            },
            "title": "Dimensions",
            "data_key": "data",
            "subheader_key": "totalCount",
        },
    },
    "metrics>list>segment": {
        "table_params": {
            "output_keys": {
                "name": {
                    "label": "Name",
                },
                "cube": {
                    "label": "Cube",
                },
                "description": {
                    "label": "Description",
                },
                "collectionId": {
                    "label": "Collection ID",
                },
                "publicationId": {
                    "label": "Publication ID",
                },
                "metadata": {
                    "label": "Metadata",
                },
            },
            "title": "Segments",
            "data_key": "data",
            "subheader_key": "totalCount",
        },
    },
    "metrics>list-collections": {
        "table_params": {
            "output_keys": {
                "id": {
                    "label": "ID",
                },
                "name": {
                    "label": "Name",
                },
                "description": {
                    "label": "Description",
                },
                "scope": {
                    "label": "Scope",
                },
                "createdAt": {
                    "label": "Created At",
                },
                "createdBy": {
                    "label": "Created By",
                },
            },
            "title": "Collections",
            "data_key": "collections",
            "subheader_key": "totalCount",
        },
    },
    "metrics>list-namespaces": {
        "table_params": {
            "output_keys": {
                "name": {
                    "label": "Name",
                },
                "models": {
                    "label": "Models",
                },
            },
            "title": "Namespaces",
            "data_key": "namespaces",
            "subheader_key": "totalCount",
        },
    },
    "artifacts>create": {
        "request_body_yaml_path": "sample_yaml/resources/artifacts/create_artifact.yaml",
    },
    "artifacts>update-metadata": {
        "request_body_yaml_path": "sample_yaml/resources/artifacts/update_artifact_metadata.yaml",
    },
    "artifacts>create-version": {
        "request_body_yaml_path": "sample_yaml/resources/artifacts/create_artifact_version.yaml",
    },
    "images>create": {
        "request_body_yaml_path": "sample_yaml/resources/images/upload/create_image.yaml",
    },
    "images>create-version": {
        "request_body_yaml_path": "sample_yaml/resources/images/upload/create_image_version.yaml",
    },
    "images>update-version": {
        "request_body_yaml_path": "sample_yaml/resources/images/upload/update_version.yaml",
    },
    "images>create-or-update": {
        "request_body_yaml_path": "sample_yaml/resources/images/upload/create_or_update_image.yaml",
    },
    "workflows>create": {
        "request_body_yaml_path": "sample_yaml/resources/workflows/create_workflow.yaml",
    },
    "workflows>update": {
        "request_body_yaml_path": "sample_yaml/resources/workflows/update_workflow.yaml",
    },
    "workflows>create-or-update": {
        "request_body_yaml_path": "sample_yaml/resources/workflows/create_or_update_workflow.yaml",
    },
    "workflows>patch": {
        "request_body_yaml_path": "sample_yaml/resources/workflows/patch_workflow.yaml",
    },
    "workflows>execute": {
        "request_body_yaml_path": "sample_yaml/resources/workflows/execute_workflow.yaml",
    },
    "webapps>create": {
        "request_body_yaml_path": "sample_yaml/resources/webapps/create_webapp.yaml",
    },
    "webapps>update": {
        "request_body_yaml_path": "sample_yaml/resources/webapps/update_webapp.yaml",
    },
    "webapps>create-or-update": {
        "request_body_yaml_path": "sample_yaml/resources/webapps/create_or_update_webapp.yaml",
    },
    "services>create": {
        "request_body_yaml_path": "sample_yaml/resources/services/create_service.yaml",
    },
    "services>update": {
        "request_body_yaml_path": "sample_yaml/resources/services/update_service.yaml",
    },
    "services>create-or-update": {
        "request_body_yaml_path": "sample_yaml/resources/services/create_or_update_service.yaml",
    },
    "services>test": {
        "request_body_yaml_path": "sample_yaml/resources/services/test_service.yaml",
    },
    "alerts>emails>send": {
        "request_body_yaml_path": "sample_yaml/resources/emails/send_email.yaml",
    },
    "apps>specs>create": {
        "request_body_yaml_path": "sample_yaml/press/apps/specs/create_app_spec.yaml",
    },
    "apps>specs>update-metadata": {
        "request_body_yaml_path": "sample_yaml/press/apps/specs/update_app_spec_metadata.yaml",
    },
    "apps>specs>create-release": {
        "request_body_yaml_path": "sample_yaml/press/apps/specs/create_app_spec_release.yaml",
    },
    "apps>deployments>create": {
        "request_body_yaml_path": "sample_yaml/press/apps/deployments/create_app_deployment.yaml",
    },
    "apps>deployments>create-revision": {
        "request_body_yaml_path": "sample_yaml/press/apps/deployments/create_app_deployment_revision.yaml",
    },
    "apps>deployments>update-metadata": {
        "request_body_yaml_path": "sample_yaml/press/apps/deployments/update_app_deployment_metadata.yaml",
    },
    "blocks>specs>create": {
        "request_body_yaml_path": "sample_yaml/press/blocks/specs/workflow/create_block_spec.yaml",
    },
    "blocks>specs>update-metadata": {
        "request_body_yaml_path": "sample_yaml/press/blocks/specs/update_block_spec_metadata.yaml",
    },
    "blocks>specs>create-release": {
        "request_body_yaml_path": "sample_yaml/press/blocks/specs/workflow/create_block_spec_release.yaml",
    },
    "blocks>deployments>create": {
        "request_body_yaml_path": "sample_yaml/press/blocks/deployments/create_block_deployment.yaml",
    },
    "blocks>deployments>create-revision": {
        "request_body_yaml_path": "sample_yaml/press/blocks/deployments/create_block_deployment_revision.yaml",
    },
    "blocks>deployments>update-metadata": {
        "request_body_yaml_path": "sample_yaml/press/blocks/deployments/update_block_deployment_metadata.yaml",
    },
    "blocks>deployments>patch-parameters": {
        "request_body_yaml_path": "sample_yaml/press/blocks/deployments/patch_block_parameters.yaml",
    },
    "metrics>publish": {
        "request_body_yaml_path": "sample_yaml/metrics/publish.yaml",
    },
    "metrics>query": {
        "request_body_yaml_path": "sample_yaml/metrics/query.yaml",
        "table_params": {
            "output_keys": {},
            "title": "Query Data",
            "data_key": "data",
            "subheader_key": "totalCount",
        },
    },
    "metrics>create-collection": {
        "request_body_yaml_path": "sample_yaml/metrics/create_collection.yaml",
    },
}

command_parameter: Dict[str, str] = {
    "metrics>list": "type",
}

__all__: List[str] = ["command_metadata", "command_parameter"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Workflow client module."""

from __future__ import annotations

from typing import Any, Dict, Iterator, List, Literal, Optional, overload

from peak.base_client import BaseClient
from peak.constants import ContentType, HttpMethods
from peak.exceptions import InvalidParameterException
from peak.helpers import (
    combine_dictionaries,
    download_logs_helper,
    map_user_options,
    variables_to_dict,
)
from peak.session import Session


class Workflow(BaseClient):
    """Client class for interacting with workflows resource."""

    BASE_ENDPOINT = "workflows/api/v1"

    @overload
    def list_workflows(
        self: Workflow,
        workflow_status: Optional[List[str]] = None,
        last_execution_status: Optional[List[str]] = None,
        last_modified_by: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_workflows(
        self: Workflow,
        workflow_status: Optional[List[str]] = None,
        last_execution_status: Optional[List[str]] = None,
        last_modified_by: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_workflows(
        self: Workflow,
        workflow_status: Optional[List[str]] = None,
        last_execution_status: Optional[List[str]] = None,
        last_modified_by: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        name: Optional[str] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve the list of workflows.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/get-workflows>`__

        Args:
            workflow_status (List[str] | None): List of status to filter workflow. Default is None.
                Valid values are `Draft`, `Running`, `Available`, `Paused`.
            last_execution_status (str | None): The last execution status of the workflow. Default is None.
                Valid values are `Success`, `Running`, `Stopped`, `Stopping`, `Failed`.
            last_modified_by (str | None): The user who last modified the workflow. Default is None.
            page_size (int | None): The number of workflows per page.
            page_number (int | None): The page number to retrieve. Only used when return_iterator is False.
            name (str | None): Search workflows by name.
            return_iterator (bool): Whether to return an iterator object or list of workflows for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: An iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
            StopIteration: There are no more pages to list
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/workflows/"
        params: Dict[str, Any] = {
            "pageSize": page_size,
            "workflowStatus": workflow_status,
            "lastExecutionStatus": last_execution_status,
            "lastModifiedBy": last_modified_by,
            "searchTerm": name,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="workflows",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def create_workflow(self: Workflow, body: Dict[str, Any]) -> Dict[str, int]:
        """Create a new workflow.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/create-workflow>`__

        Args:
            body (Dict[str, Any]): A dictionary containing the workflow config. Schema can be found below.

        Returns:
            Dict[str, int]: Id of the newly created workflow.

        SCHEMA:
            .. code-block:: json

                {
                    "name": "string(required)",
                    "triggers": [
                        {
                            "cron": "string"
                        },
                        {
                            "webhook": "boolean",
                        }
                    ],
                    "watchers": [
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "user": "string",
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "webhook": {
                                "name": "string(required)",
                                "url": "string(required)",
                                "payload": "string(required)"
                            }
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "email": {
                                "name": "string",
                                "recipients": {
                                    "to": ["string"]
                                }
                            }
                        }
                    ],
                    "retryOptions": {
                        "duration": "number",
                        "exitCodes": [],
                        "exponentialBackoff": "boolean",
                        "numberOfRetries": "number"
                    },
                    "tags": [
                        {
                            "name": "string"
                        },
                        {
                            "name": "string"
                        }
                    ],
                    "steps": {
                        "stepName": {
                            "type": "standard",
                            "imageId": "number",
                            "imageVersionId": "number",
                            "command": "string",
                            "resources": {
                                "instanceTypeId": "number",
                                "storage": "string"
                            },
                            "parents": [],
                            "stepTimeout": "number",
                            "clearImageCache": "boolean",
                            "repository": {
                                "branch": "string",
                                "token": "string",
                                "url": "string"
                            },
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": [],
                            },
                            "outputParameters": {
                                "key (string)": "value (string)"
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName2": {
                            "type": "http",
                            "method": "enum(get, post, put, patch, delete)",
                            "url": "string",
                            "payload": "string",
                            "headers": {
                                "absolute": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": {
                                    "key (string)": "value (string)"
                                }
                            },
                            "auth": {
                                "type": "enum(no-auth, oauth, basic, api-key, bearer-token)(required)",
                                "clientId": "string | required for oauth",
                                "clientSecret": "string | required for oauth",
                                "authUrl": "string | required for oauth",
                                "username": "string | required for basic",
                                "password": "string | required for basic",
                                "apiKey": "string | required for api-key",
                                "bearerToken": "string | required for bearer-token"
                            },
                            "parents": [],
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": [],
                            },
                            "outputParameters": {
                                "key (string)": "value (string)"
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName3": {
                            "type": "export",
                            "schema": "string (required)",
                            "table": "string (required)",
                            "sortBy": "string (required)",
                            "sortOrder": "enum(asc, desc) | Default is asc",
                            "compression": "boolean | Default is false",
                            "parents": [],
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName4": {
                            "type": "sql",
                            "sqlQueryPath": "string",
                            "repository": {
                                "branch": "string | Required if repository is provided",
                                "token": "string",
                                "url": "string | Required if repository is provided",
                                "filePath": "string | Required if repository is provided"
                            }
                            "parents": [],
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                }
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        }
                    }
                }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/workflows/"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
        )

    def create_or_update_workflow(self: Workflow, body: Dict[str, Any]) -> Dict[str, int]:
        """Creates a new workflow or updates an existing workflow based on workflow name.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/create-workflow>`__

        Args:
            body (Dict[str, Any]): A dictionary containing the workflow config. Schema can be found below.

        Returns:
            Dict[str, int]: Id of the newly created or updated workflow.

        SCHEMA:
            .. code-block:: json

                {
                    "name": "string(required)",
                    "triggers": [
                        {
                            "cron": "string"
                        },
                        {
                            "webhook": "boolean",
                        }
                    ],
                    "watchers": [
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "user": "string",
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "webhook": {
                                "name": "string(required)",
                                "url": "string(required)",
                                "payload": "string(required)"
                            }
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "email": {
                                "name": "string",
                                "recipients": {
                                    "to": ["string"]
                                }
                            }
                        }
                    ],
                    "retryOptions": {
                        "duration": "number",
                        "exitCodes": [],
                        "exponentialBackoff": "boolean",
                        "numberOfRetries": "number"
                    },
                    "tags": [
                        {
                            "name": "string"
                        },
                        {
                            "name": "string"
                        }
                    ],
                    "steps": {
                        "stepName": {
                            "type": "http",
                            "imageId": "number",
                            "imageVersionId": "number",
                            "command": "string",
                            "resources": {
                                "instanceTypeId": "number",
                                "storage": "string"
                            },
                            "parents": [],
                            "stepTimeout": "number",
                            "clearImageCache": "boolean",
                            "repository": {
                                "branch": "string",
                                "token": "string",
                                "url": "string"
                            },
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": [],
                            },
                            "outputParameters": {
                                "key (string)": "value (string)"
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName2": {
                            "type": "http",
                            "method": "enum(get, post, put, patch, delete)(required)",
                            "url": "string(required)",
                            "payload": "string",
                            "headers": {
                                "absolute": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": {
                                    "key (string)": "value (string)"
                                }
                            },
                            "auth": {
                                "type": "enum(no-auth, oauth, basic, api-key, bearer-token)",
                                "clientId": "string | required for oauth",
                                "clientSecret": "string | required for oauth",
                                "authUrl": "string | required for oauth",
                                "username": "string | required for basic",
                                "password": "string | required for basic",
                                "apiKey": "string | required for api-key",
                                "bearerToken": "string | required for bearer-token"
                            },
                            "parents": [],
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": [],
                            },
                            "outputParameters": {
                                "key (string)": "value (string)"
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName3": {
                            "type": "export",
                            "schema": "string (required)",
                            "table": "string (required)",
                            "sortBy": "string (required)",
                            "sortOrder": "enum(asc, desc) | Default is asc",
                            "compression": "boolean | Default is false",
                            "parents": [],
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName4": {
                            "type": "sql",
                            "sqlQueryPath": "string",
                            "repository": {
                                "branch": "string | Required if repository is provided",
                                "token": "string",
                                "url": "string | Required if repository is provided",
                                "filePath": "string | Required if repository is provided"
                            }
                            "parents": [],
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                }
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        }
                    }
                }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        workflow_name = body.get("name", "")
        response = (
            {}
            if not len(workflow_name)
            else self.list_workflows(page_size=100, return_iterator=False, name=workflow_name)
        )
        filtered_workflows = list(
            filter(lambda workflow: workflow.get("name", "") == workflow_name, response.get("workflows", [])),
        )

        if len(filtered_workflows) > 0:
            workflow_id = filtered_workflows[0]["id"]
            return self.update_workflow(workflow_id=workflow_id, body=body)

        return self.create_workflow(body=body)

    def describe_workflow(
        self: Workflow,
        workflow_id: int,
    ) -> Dict[str, Any]:
        """Retrieve details of a specific workflow.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/describe-workflow>`__

        Args:
            workflow_id (int): The ID of the workflow to retrieve.

        Returns:
            Dict[str, Any]: A dictionary containing the details of the workflow.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/workflows/{workflow_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def update_workflow(self: Workflow, workflow_id: int, body: Dict[str, Any]) -> Dict[str, Any]:
        """Update an existing workflow.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/update-workflow>`__

        Args:
            workflow_id (int): The ID of the workflow to update.
            body (dict): A dictionary containing the updated workflow details. Schema can be found below.

        Returns:
            Dict[str, int]: Id of the updated workflow.

        SCHEMA:
            .. code-block:: json

                {
                    "name": "string",
                    "triggers": [
                        {
                            "cron": "string"
                        },
                        {
                            "webhook": "boolean",
                        }
                    ],
                    "watchers": [
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "user": "string",
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "webhook": {
                                "name": "string(required)",
                                "url": "string(required)",
                                "payload": "string(required)"
                            }
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "email": {
                                "name": "string",
                                "recipients": {
                                    "to": ["string"]
                                }
                            }
                        }
                    ],
                    "retryOptions": {
                        "duration": "number",
                        "exitCodes": [],
                        "exponentialBackoff": "boolean",
                        "numberOfRetries": "number"
                    },
                    "tags": [
                        {
                            "name": "string"
                        },
                        {
                            "name": "string"
                        }
                    ],
                    "steps": {
                        "stepName": {
                            "type": "standard",
                            "imageId": "number",
                            "imageVersionId": "number",
                            "command": "string",
                            "resources": {
                                "instanceTypeId": "number",
                                "storage": "string"
                            },
                            "parents": [],
                            "stepTimeout": "number",
                            "clearImageCache": "boolean",
                            "repository": {
                                "branch": "string",
                                "token": "string",
                                "url": "string"
                            },
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": [],
                            },
                            "outputParameters": {
                                "key (string)": "value (string)"
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName2": {
                            "type": "http",
                            "method": "enum(get, post, put, patch, delete)(required)",
                            "url": "string(required)",
                            "payload": "string",
                            "headers": {
                                "absolute": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": {
                                    "key (string)": "value (string)"
                                }
                            },
                            "auth": {
                                "type": "enum(no-auth, oauth, basic, api-key, bearer-token)(required)",
                                "clientId": "string | required for oauth",
                                "clientSecret": "string | required for oauth",
                                "authUrl": "string | required for oauth",
                                "username": "string | required for basic",
                                "password": "string | required for basic",
                                "apiKey": "string | required for api-key",
                                "bearerToken": "string | required for bearer-token"
                            },
                            "parents": [],
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": [],
                            },
                            "outputParameters": {
                                "key (string)": "value (string)"
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName3": {
                            "type": "export",
                            "schema": "string (required)",
                            "table": "string (required)",
                            "sortBy": "string (required)",
                            "sortOrder": "enum(asc, desc) | Default is asc",
                            "compression": "boolean | Default is false",
                            "parents": [],
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        },
                        "stepName4": {
                            "type": "sql",
                            "sqlQueryPath": "string",
                            "repository": {
                                "branch": "string | Required if repository is provided",
                                "token": "string",
                                "url": "string | Required if repository is provided",
                                "filePath": "string | Required if repository is provided"
                            }
                            "parents": [],
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                }
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            }
                        }
                    }
                }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            ConflictException: The workflow is in a conflicting state while deleting.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.PUT, f"{self.BASE_ENDPOINT}/workflows/{workflow_id}"
        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
        )

    def patch_workflow(  # noqa: C901, PLR0912
        self: Workflow,
        workflow_id: int,
        body: Optional[Dict[str, Any]] = None,
        name: Optional[str] = None,
        repository: Optional[str] = None,
        branch: Optional[str] = None,
        token: Optional[str] = None,
        command: Optional[str] = None,
        image_id: Optional[int] = None,
        image_version_id: Optional[int] = None,
        instance_type_id: Optional[int] = None,
        storage: Optional[str] = None,
        step_timeout: Optional[int] = None,
        clear_image_cache: Optional[bool] = None,
        step_names: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """Update an existing workflow.

        - This function allows to efficiently modify trigger details, watchers, workflow name, and specific step attributes such as repository URL, branch, token, image ID, version ID etc.

        - By specifying `step_names`, we can globally update specified steps with provided parameters, streamlining the update process. If `step_names` is not provided, all the steps for that workflow would be updated.

        - Alternatively, we can utilize the **body** parameter to selectively modify individual step attributes across different steps. With this, we can also add new steps to the workflow by providing the parameters required by the step.

        - If both body and specific parameters are used, the latter takes precedence.

        Args:
            workflow_id (int): The ID of the workflow to patch.
            body (Dict[str, Any] | None): A dictionary containing the updated workflow details.
            name (str | None): The name of the workflow.
            repository (str | None): URL of the repository containing the required files.
            branch (str | None): The branch of the repository to use.
            token (str | None): The token to be used to access the repository.
            command (str | None): The command to run when workflow step is executed.
            image_id (int | None): The ID of the image to use for the workflow step.
            image_version_id (int | None): The ID of the image version to use for the workflow step.
            instance_type_id (int | None): The ID of the instance type to use for the workflow step.
            storage (str | None): The storage to use for the workflow step in GB. For example, "10GB".
            step_timeout (int | None): Time after which the step timeouts.
            clear_image_cache (boolean | None): Whether to clear image cache on workflow execution.
            step_names (List[str] | None): The workflow steps to update. If not provided, all steps will be updated.

        SCHEMA:
            .. code-block:: json

                {
                    "name": "string",
                    "triggers": [
                        {
                            "cron": "string"
                        },
                        {
                            "webhook": "boolean",
                        }
                    ],
                    "watchers": [
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "user": "string",
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "webhook": {
                                "name": "string(required)",
                                "url": "string(required)",
                                "payload": "string(required)"
                            }
                        },
                        {
                            "events": {
                                "success": "boolean",
                                "fail": "boolean",
                                "runtimeExceeded": "number"
                            },
                            "email": {
                                "name": "string",
                                "recipients": {
                                    "to": ["string"]
                                }
                            }
                        }
                    ],
                    "retryOptions": {
                        "duration": "number",
                        "exitCodes": [],
                        "exponentialBackoff": "boolean",
                        "numberOfRetries": "number"
                    },
                    "tags": [
                        {
                            "name": "string"
                        },
                        {
                            "name": "string"
                        }
                    ],
                    "steps": {
                        "stepName": {
                            "type": "standard",
                            "imageId": "number",
                            "imageVersionId": "number",
                            "command": "string",
                            "resources": {
                                "instanceTypeId": "number",
                                "storage": "string"
                            },
                            "parents": [],
                            "stepTimeout": "number",
                            "clearImageCache": "boolean",
                            "repository": {
                                "branch": "string",
                                "token": "string",
                                "url": "string"
                            },
                            "runConfiguration": {
                                "retryOptions": {
                                    "duration": "number",
                                    "exitCodes": [],
                                    "exponentialBackoff": "boolean",
                                    "numberOfRetries": "number"
                                },
                                "skipConfiguration": {
                                    "skip": "boolean",
                                    "skipDAG": "boolean",
                                }
                            },
                            "parameters": {
                                "env": {
                                    "key (string)": "value (string)"
                                },
                                "inherit": {
                                    "key (string)": "value (string)"
                                },
                                "secrets": [],
                            },
                            "outputParameters": {
                                "key (string)": "value (string)"
                            },
                            "executionParameters": {
                                "conditional": [
                                    {
                                        "condition": "string",
                                        "paramName": "string",
                                        "stepName": "string",
                                        "value": "string"
                                    }
                                ],
                                "parentStatus": [
                                    {
                                        "condition": "string",
                                        "parents": [],
                                        "status": []
                                    }
                                ]
                            }
                        }
                    }
                }

        Returns:
            Dict[str, Any]: A dictionary containing the workflow ID.

        Raises:
            InvalidParameterException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            ConflictException: The workflow is in a conflicting state while deleting.
            InternalServerErrorException: The server failed to process the request.
        """
        user_options = variables_to_dict(
            name,
            repository,
            branch,
            token,
            command,
            image_id,
            image_version_id,
            instance_type_id,
            storage,
            step_timeout,
            clear_image_cache,
        )

        if not user_options and not body:
            raise InvalidParameterException(
                message="Request body or at least one parameter must be provided to update the workflow.",
            )

        workflow_details = self.describe_workflow(workflow_id=workflow_id)

        for step_name in step_names or []:
            if step_name not in workflow_details["steps"]:
                raise InvalidParameterException(
                    message=f"Step name {step_name} does not exist for workflow {workflow_id}",
                )

        # First, we modify steps in workflow details to the expected format for the API

        step_keys_to_skip = ["imageName", "imageVersion"]

        updated_workflow_details: Dict[str, Any] = {
            "name": workflow_details["name"],
            "tags": workflow_details["tags"],
            "watchers": workflow_details["watchers"],
            "triggers": [],
            "steps": {},
        }

        if workflow_details.get("triggers"):
            triggers = workflow_details["triggers"]
            if triggers[0].get("cron"):
                updated_workflow_details["triggers"] = [{"cron": triggers[0]["cron"]}]
            elif triggers[0].get("webhook"):
                updated_workflow_details["triggers"] = [{"webhook": True, "webhookPolicy": "preserve"}]

        for step_name, step in workflow_details["steps"].items():
            updated_workflow_details["steps"][step_name] = {}
            for key, value in step.items():
                if key not in step_keys_to_skip and value is not None:
                    updated_workflow_details["steps"][step_name][key] = value
                if key == "repository" and not value.get("token", {}):
                    updated_workflow_details["steps"][step_name]["repository"].pop("token", None)
                if key == "parameters" and "env" in value:  # Describe workflow returns env as a list of dictionaries
                    env = {}
                    for env_dict in value["env"]:
                        env.update(env_dict)
                    updated_workflow_details["steps"][step_name]["parameters"]["env"] = env

        # Second, we merge updated_workflow_details with the user provided body

        updated_body: Dict[str, Any] = updated_workflow_details.copy()

        if body:
            for key, value in body.items():
                if key == "steps":
                    for step_name, step in body["steps"].items():
                        updated_body["steps"][step_name] = combine_dictionaries(
                            updated_body["steps"].get(step_name, {}),
                            step,
                            nested_keys_to_skip=["env"],
                        )
                else:
                    updated_body[key] = value

        # Finally, we update the body with the user provided parameters

        if user_options.get("repository"):
            user_options["url"] = user_options["repository"]
            del user_options["repository"]

        keys_mapping: Dict[str, str] = {
            "url": "repository",
            "branch": "repository",
            "token": "repository",
            "instanceTypeId": "resources",
            "storage": "resources",
        }

        user_options = map_user_options(user_options, keys_mapping)

        if user_options:
            if user_options.get("name"):
                updated_body["name"] = user_options["name"]
                del user_options["name"]
            for step_name in updated_body["steps"]:
                if not step_names or step_name in step_names:
                    updated_body["steps"][step_name] = combine_dictionaries(
                        updated_body["steps"].get(step_name, {}),
                        user_options,
                    )

        return self.update_workflow(workflow_id=workflow_id, body=updated_body)

    def delete_workflow(
        self: Workflow,
        workflow_id: int,
    ) -> Dict[None, None]:
        """Delete a workflow.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/delete-workflow>`__

        Args:
            workflow_id (int): The ID of the workflow to delete.

        Returns:
            dict: Empty dictionary object.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            ConflictException: If the workflow is in a conflicting state while deleting.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT}/workflows/{workflow_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def execute_workflow(
        self: Workflow,
        workflow_id: int,
        body: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, str]:
        """Start a workflow run.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/execute-workflow>`__

        Args:
            workflow_id (int): ID of the workflow to delete.
            body: (Dict[str, Any]): The parameters to be passed while running the workflow. More details can be found in the API doc - https://service.peak.ai/workflows/api-docs/index.htm#/Workflows/execute-workflow

        Returns:
            Dict[str, str]: Execution ID of the run.

        SCHEMA:
            .. code-block:: json

                {
                    "params": {
                        "global": {
                            "key (string)": "value (string)"
                        },
                        "stepWise": {
                            "stepName (string)": {
                                "key (string)": "value (string)"
                            }
                        }
                    },
                    "stepsToRun": [
                        {
                            "stepName": "string",
                            "runDAG": "boolean"
                        }
                    ],
                }

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            ConflictException: The workflow is in a conflicting state and new run cannot be started.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/workflows/{workflow_id}/execute"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
        )

    def list_resources(
        self: Workflow,
    ) -> List[Dict[str, Any]]:
        """Lists all available resources for the workflows.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Resources/get-resources>`__

        Returns:
            Dict[str, Any]: A dictionary containing the list of available resources.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            ConflictException: The workflow is in a conflicting state and new run cannot be started.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/resources"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def get_default_resource(
        self: Workflow,
    ) -> Dict[str, Any]:
        """Default resource values that will be used in case `resource` key is not provided for the workflows.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Resources/get-default-resources>`__

        Returns:
            Dict[str, Any]: Default resource values

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/resources/defaults"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    @overload
    def list_executions(
        self: Workflow,
        workflow_id: int,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        status: Optional[List[str]] = None,
        count: Optional[int] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_executions(
        self: Workflow,
        workflow_id: int,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        status: Optional[List[str]] = None,
        count: Optional[int] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_executions(
        self: Workflow,
        workflow_id: int,
        date_from: Optional[str] = None,
        date_to: Optional[str] = None,
        status: Optional[List[str]] = None,
        count: Optional[int] = None,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Lists executions for the given workflow.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Executions/get-workflow-executions>`__

        Args:
            workflow_id (int): ID of the workflow to fetch executions.
            date_from (str | None): The date after which the executions should be included (in ISO format). Defaults to None
            date_to (str | None): The date till which the executions should be included (in ISO format). Defaults to None
            status (List[str] | None): The status of the executions to filter by.
                Valid values are `Success`, `Running`, `Stopped`, `Stopping` and `Failed`.
            count (int | None): Number of executions required in the provided time range or 90 days (Ordered by latest to earliest).
                For example, if 5 is provided, it will return last 5 workflow executions.
                By default, it will return all the executions.
            page_size (int | None): Number of executions per page.
            page_number (int | None): Page number to fetch. Only used when return_iterator is False.
            return_iterator (bool): Whether to return an iterator object or list of executions for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any]] | Dict[str, Any]: An iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.

            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            InternalServerErrorException: The server failed to process the request.
            StopIteration: There are no more pages to list
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/workflows/executions/{workflow_id}"

        params: Dict[str, Any] = {
            "count": count,
            "statuses": status,
            "dateTo": date_to,
            "dateFrom": date_from,
            "pageSize": page_size,
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="executions",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def get_execution_logs(
        self: Workflow,
        workflow_id: int,
        execution_id: str,
        step_name: str,
        next_token: Optional[str] = None,
        save: Optional[bool] = False,  # noqa: FBT002
        file_name: Optional[str | None] = None,
    ) -> Dict[str, Any] | None:
        """Get workflow execution logs.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Executions/get-workflow-step-logs>`__

        Args:
            workflow_id (int): ID of the workflow to fetch executions.
            execution_id (str): ID of the execution to fetch logs.
            step_name (str): Name of the step to fetch logs.
            next_token (str | None): The token to retrieve the next set of logs. Defaults to None.
            save (bool): Whether to save the logs to a file. Defaults to False.
            file_name (str | None): File name or path where the contents should be saved. Default file name is `workflow_execution_logs_<workflow_id>_<execution_id>_<step_name>.log`.

        Returns:
            Dict[str, Any]: A dictionary containing the logs for the given execution.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method = HttpMethods.GET
        get_logs_endpoint = f"{self.BASE_ENDPOINT}/workflows/{workflow_id}/executions/{execution_id}/logs"
        download_logs_endpoint = f"{get_logs_endpoint}/download"

        if save:
            response = self.session.create_request(
                download_logs_endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                params={"stepName": step_name},
            )
            updated_file_name = (
                f"workflow_execution_logs_{workflow_id}_{execution_id}_{step_name}.log"
                if not file_name or not len(file_name)
                else file_name
            )
            download_logs_helper(response=response, file_name=updated_file_name)
            return None

        params: Dict[str, Any] = {
            "stepName": step_name,
            "nextToken": next_token,
        }

        return self.session.create_request(  # type: ignore[no-any-return]
            get_logs_endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params=params,
        )

    def get_execution_details(
        self: Workflow,
        workflow_id: int,
        execution_id: str,
    ) -> Dict[str, Any] | None:
        """Get details about an execution.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/workflows/api-docs/index.htm#/Executions/get-workflow-execution-details>`__

        Args:
            workflow_id (int): ID of the workflow to which the execution belongs.
            execution_id (str): ID of the execution to get details for.

        Returns:
            Dict[str, Any]: A dictionary containing the logs for the given execution.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given workflow does not exist.
            InternalServerErrorException: The server failed to process the request.
        """
        method = HttpMethods.GET
        get_details_endpoint = f"{self.BASE_ENDPOINT}/workflows/{workflow_id}/executions/{execution_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            get_details_endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )


def get_client(session: Optional[Session] = None) -> Workflow:
    """Returns a Workflow client, If no session is provided, a default session is used.

    Args:
        session (Optional[Session]): A Session Object. Default is None.

    Returns:
        Workflow: the workflow client object
    """
    return Workflow(session)


__all__: List[str] = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Generic utilities."""

from __future__ import annotations

import re
from typing import Any, MutableMapping

# Masking patterns
BANK_CARD_PATTERN: re.Pattern[str] = re.compile(r"\b(?:\d[- ]?){13,16}\b")
EMAIL_PATTERN: re.Pattern[str] = re.compile(r"[\w+\.+\-]+@+[\w+\.+\-]+[\.\w]{2,}")
PHONE_NUM_PATTERN: re.Pattern[str] = re.compile(
    r"\+\d{1,3}\s?(\(\d{1,4}\)|\d{1,4})[\s.-]?\d{1,4}[\s.-]?\d{1,9}",
)
JWT_PATTERN: re.Pattern[str] = re.compile(r"\bey[a-zA-Z0-9_-]*\.[a-zA-Z0-9_-]*\.[a-zA-Z0-9_-]*\b")
SSN_PATTERN: re.Pattern[str] = re.compile(r"\b(\d{4}[-\s]?\d{4}[-\s]?\d{4}|\d{3}-\d{2}-\d{4})\b")


# Masking configs
MASK_SYMBOL = "*"
MAX_CHARS_TO_MASK = 32
VISIBLE_CHARS_FROM_START = 3
VISIBLE_CHARS_FROM_END = 3

PII_PATTERNS: list[re.Pattern[str]] = [
    BANK_CARD_PATTERN,
    EMAIL_PATTERN,
    PHONE_NUM_PATTERN,
    JWT_PATTERN,
    SSN_PATTERN,
]


def mask_pii_data(_str: str) -> str:
    """Masks sensitive PII data in the input string.

    Args:
        _str (str): The input string containing sensitive information.

    Returns:
        str: The same input string with sensitive information masked.
    """

    def mask_pattern(match: re.Match[str]) -> str:
        match_text: str = match.group()
        mask_length: int = min(len(match_text), MAX_CHARS_TO_MASK)
        visible_start: str = match_text[:VISIBLE_CHARS_FROM_START]
        visible_end: str = match_text[-VISIBLE_CHARS_FROM_END:]
        masked_chars: str = MASK_SYMBOL * mask_length
        return visible_start + masked_chars + visible_end

    masked_string: str = _str
    for regex_pattern in PII_PATTERNS:
        masked_string = regex_pattern.sub(mask_pattern, masked_string)

    return masked_string


def mask_nested_pii_data(_dict: MutableMapping[str, Any]) -> MutableMapping[str, Any]:
    """Masks sensitive PII data in the input dictionary.

    Args:
        _dict (MutableMapping): The input dictionary containing sensitive information.

    Returns:
        MutableMapping: The same input dictionary with sensitive information masked.
    """

    def mask_value(value: Any) -> Any:
        if isinstance(value, str):
            return mask_pii_data(value)
        return value

    for key, value in _dict.items():
        if isinstance(value, MutableMapping):
            # Recursively call mask_nested_pii_data on nested mappings
            _dict[key] = mask_nested_pii_data(value)
        else:
            _dict[key] = mask_value(value)

    return _dict



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Helper functions for Peak `cli`."""

from __future__ import annotations

import inspect
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional

import yaml
from peak import base_client, press, resources
from peak.exceptions import BadParameterException
from peak.helpers import format_date, remove_none_values
from peak.metrics import metrics
from peak.template import load_template
from rich.console import Console

console = Console()


def parse_params(params: Optional[List[str]]) -> Dict[str, str]:
    """Parse parameters into key-value pairs.

    Args:
        params (Optional[List[str]]): List of key-value pairs

    Returns:
        Dict[str, str]: Dictionary of key-value pairs

    Raises:
        BadParameterException: If a value is invalid

    >>> params = ["foo=1", "bar=2"]
    >>> vals = parse_params(params)
    >>> vals
    {'foo': '1', 'bar': '2'}

    Raises an error if a value is invalid
    >>> parse_params(["bar==foo"]) #doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    BadParameterException: Unable to parse: bar==foo
    """
    if params is None:
        return {}

    _to_pass: Dict[str, Any] = {}
    for keyval in params:
        if keyval.count("=") != 1:
            raise BadParameterException(keyval)
        s: List[str] = keyval.split("=", 1)  # split on first '='
        _to_pass[s[0]] = s[-1]
    return _to_pass


def check_file_extension(file_path: str) -> None:
    """Checks if the file has a .txt or .md extension.

    Args:
        file_path (str): Path to the file to check.

    Raises:
        ValueError: If the file extension is not .txt or .md.
    """
    file_type = Path(file_path).suffix
    if file_type not in [".txt", ".md"]:
        msg = f"Unsupported file type: `{file_type}` Only .txt and .md files are supported."
        raise ValueError(msg)


def template_handler(
    file: str,
    params_file: Optional[str] = None,
    params: Optional[List[str]] = None,
    markdown_files: Optional[Dict[str, str]] = None,
) -> Dict[str, Any]:
    """Loads and returns the rendered template.

    Args:
        file (str): Path to the template file.
        params_file (Optional[str]): Path to the params map file.
        params (Optional[List[str]]): List of params to override.
        markdown_files (Optional[Dict[str, str]]): Dictionary of markdown files to load.
            The key is a colon-separated string representing the nested key path (e.g., "body:metadata:description"), and the value is the path to the markdown file.

    Returns:
        Dict[str, Any]: Rendered template with values substituted.
    """
    params_dict: Dict[str, Any] = {}
    markdown_data: Dict[str, str] = {}

    if params_file:
        with Path(params_file).open("r") as f:
            params_dict = yaml.safe_load(f.read())

    if markdown_files:
        for key, value in markdown_files.items():
            check_file_extension(value)
            with Path(value).open("r") as f:
                markdown_data[key] = f.read()

    params_dict = remove_none_values({**params_dict, **parse_params(params)})

    return load_template(file, params_dict, markdown_data=markdown_data, convert_to_snake_case=True)


def remove_unknown_args(args: Dict[str, Any], func: Callable[..., Any]) -> Dict[str, Any]:
    """Filters keys from dictionary which are not accepted by the function.

    Args:
        args (Dict[str, Any]): dictionary to filter
        func (Callable): function to filter for

    Returns:
        Dict[str, Any]: filtered dictionary
    """
    func_params = inspect.signature(func).parameters
    return {k: v for k, v in args.items() if k in func_params}


def get_updated_artifacts(
    body: Dict[str, Any],
    artifact_path: str | None,
    artifact_ignore_files: List[str] | None,
) -> Dict[str, Any]:
    """Returns updated artifacts by replacing artifact path and ignore files by the ones provided via cli args.

    Args:
        body (Dict[str, Any]): Dictionary containing request body generated from user provided yaml file
        artifact_path (str | None): Path to the artifact.
        artifact_ignore_files (List[str] | None): Ignore files to use when creating artifact.

    Returns:
        Dict[str, Any]: Dictionary containing updated artifacts
    """
    artifact: Dict[str, Any] = {}

    if artifact_path:
        artifact["path"] = artifact_path
    elif "artifact" in body and "path" in body["artifact"]:
        artifact["path"] = body["artifact"]["path"]

    if artifact_ignore_files:
        artifact["ignore_files"] = artifact_ignore_files
    elif "artifact" in body and "ignore_files" in body["artifact"]:
        artifact["ignore_files"] = body["artifact"]["ignore_files"]

    return artifact


def parse_build_arguments(build_arguments: List[str]) -> List[Dict[str, str]]:
    """Parses build arguments provided via cli args to the format {name: arg1, value: value1}.

    Args:
        build_arguments (List[str]): List of build arguments provided via cli args

    Returns:
        List[Dict[str, str]]: List of build arguments in the required format

    Raises:
        BadParameterException: If a value is invalid
    """
    parsed_build_arguments: List[Dict[str, str]] = []
    for build_argument in build_arguments:
        try:
            key, value = build_argument.split("=", 1)
        except ValueError as err:
            raise BadParameterException(
                build_argument,
                message="Invalid build argument format. It should in the --build-arguments arg1=value1 format",
            ) from err
        key, value = build_argument.split("=", 1)
        parsed_build_arguments.append({"name": key, "value": value})
    return parsed_build_arguments


def parse_envs(env: List[str]) -> Dict[str, str]:
    """Parses envs provided via cli args to the format {arg1: value1}.

    Args:
        env (List[str]): List of envs provided via cli args.

    Returns:
        Dict[str, str]: Envs in the required format.

    Raises:
        BadParameterException: If a value is invalid.
    """
    parsed_envs: Dict[str, str] = {}
    for env_arg in env:
        try:
            key, value = env_arg.split("=", 1)
        except ValueError as err:
            raise BadParameterException(
                env_arg,
                message="Invalid env format. It should in the --env arg1=value1 format",
            ) from err
        parsed_envs[key] = value
    return parsed_envs


def format_logs(logs: List[Dict[str, Any]]) -> str:
    """Formats logs into a readable format.

    Args:
        logs (List[Dict[str, Any]]): List of logs

    Returns:
        str: Formatted logs
    """
    formatted_logs = ""
    for log in logs:
        updated_log = f"{format_date(timestamp=log['timestamp'])}: {log['message']}\n\n"
        formatted_logs += updated_log

    return formatted_logs


def get_client(command: str) -> base_client.BaseClient:
    """Create a client for the invoked command.

    Args:
        command (str): Invoked CLI command

    Returns:
        BaseClient: client class for the invoked command
    """
    command_client_map: Dict[str, ModuleType] = {
        "apps": press.apps,
        "blocks": press.blocks,
        "specs": press.specs,
        "deployments": press.deployments,
        "artifacts": resources.artifacts,
        "images": resources.images,
        "workflows": resources.workflows,
        "services": resources.services,
        "webapps": resources.webapps,
        "tenants": resources.tenants,
        "users": resources.users,
        "alerts": resources.alerts,
        "metrics": metrics,
    }
    return command_client_map[command].get_client()  # type: ignore[no-any-return]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

"""User client module."""
from __future__ import annotations

from typing import Dict, Optional

from peak.base_client import BaseClient
from peak.constants import ContentType, HttpMethods
from peak.helpers import search_action
from peak.session import Session
from peak.validators import validate_action, validate_feature_path


class User(BaseClient):
    """Client class for interacting with users resource."""

    BASE_ENDPOINT = "access-control/api/v2"

    def check_permissions(self, feature_actions: Dict[str, str]) -> Dict[str, bool]:
        """Check if the user has the specified permissions for the given features or subfeatures.

        Args:
            feature_actions (Dict[str, str]): A dictionary where keys are feature paths (e.g., "PRICING.GUARDRAILS")
                and values are actions (e.g., "read" or "write").

        Returns:
            Dict[str, bool]: A dictionary where keys are the same feature paths and values are booleans
                indicating whether the user has the specified action permission for that feature.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given feature does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/access/permissions"

        permissions_response = self.session.create_request(
            endpoint,
            method,
            params={
                "type": "app",
            },
            content_type=ContentType.APPLICATION_JSON,
        )

        result_permissions = {}

        for feature_path, action in feature_actions.items():
            validate_action(action)

            search_result = search_action(feature_path, permissions_response, action)
            validate_feature_path(search_result=search_result, feature_path=feature_path)

            result_permissions[feature_path] = search_result["has_permission"]

        return result_permissions


def get_client(session: Optional[Session] = None) -> User:
    """Returns a User client, If no session is provided, a default session is used.

    Args:
        session (Optional[Session]): A Session Object. Default is None.

    Returns:
        User: The user client object.
    """
    return User(session)


__all__ = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""All the Typer callbacks."""
from __future__ import annotations

from pathlib import Path

import click
import typer

import peak.config
from peak._metadata import command_metadata, command_parameter
from peak.constants import OutputTypes
from peak.output import Writer

writer = Writer(ignore_debug_mode=True)


def dry_run(*, debug_mode: bool) -> None:
    """Callback to enable dry run."""
    peak.config.DEBUG_MODE = debug_mode


def paging(*, enable_paging: bool) -> None:
    """Callback to enable paging."""
    peak.config.ENABLE_PAGING = enable_paging


def update_command(ctx: typer.Context, command: str) -> str:
    """Callback to update the command based on the command parameter."""
    if not ctx or not hasattr(ctx, "params") or command not in command_parameter:
        return command

    param = command_parameter[command]
    param_value = ctx.params.get(param)
    if param_value:
        command_with_params = f"{command}>{param_value}"
        if command_with_params in command_metadata and "table_params" in command_metadata[command_with_params]:
            return command_with_params
    return command


def get_full_command_name(ctx: typer.Context | click.core.Context | None) -> str:
    """Callback to get the full command name."""
    if ctx:
        parent = get_full_command_name(ctx.parent)
        if parent:
            return f"{parent}>{ctx.info_name}"
        return str(ctx.info_name)

    return ""


def get_command_name(ctx: typer.Context) -> str:
    """Callback to get the parsed command name with first part removed."""
    command = get_full_command_name(ctx)
    return ">".join(command.split(">")[1:])


def handle_output(output_type: OutputTypes, ctx: typer.Context) -> None:
    """Callback to handle CLI output type and set all the required parameters for Table output."""
    peak.config.OUTPUT_TYPE = output_type

    command = get_command_name(ctx)
    command = update_command(ctx, command)

    if (command not in command_metadata) or ("table_params" not in command_metadata[command]):
        return

    peak.config.TABLE_PARAMS = command_metadata[command]["table_params"]


def generate_yaml(ctx: typer.Context, generate_yaml: bool) -> None:  # noqa: FBT001 # pragma: no cover
    """Callback to generate yaml file."""
    command = get_command_name(ctx)
    output_type = peak.config.OUTPUT_TYPE

    if (
        (command not in command_metadata)
        or ("request_body_yaml_path" not in command_metadata[command])
        or output_type == "table"
    ):
        return

    if generate_yaml:
        current_file_path = Path(__file__).resolve().parent
        file_path = Path(f"{current_file_path}/{command_metadata[command]['request_body_yaml_path']}").resolve()
        with file_path.open() as file:
            yaml_text = file.read()
            with writer.pager():
                writer.write(yaml_text)
                raise typer.Exit



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Log Writer Module."""

from __future__ import annotations

import contextlib
import json
import sys
from typing import Any, List

import yaml
from rich.console import Console, PagerContext
from rich.highlighter import JSONHighlighter
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from peak import config
from peak.constants import OutputTypes

console = Console()
json_highlighter = JSONHighlighter()


class Writer:
    """The Writer class used to print out data in the CLI."""

    def __init__(self, *, ignore_debug_mode: bool = False) -> None:
        """Initialise the write class.

        Args:
            ignore_debug_mode (bool): don't consider debug mode and print the output anyways.
        """
        self.ignore_debug_mode = ignore_debug_mode

    def write(
        self,
        data: Any,
        deprecation_message: str | None = None,
        output_type: OutputTypes | None = None,
    ) -> None:
        """Write logs to the terminal.

        Args:
            data (Any): Data to be printed on the terminal.
                This handles dry-run, debug mode and exit code for the CLI.
            deprecation_message (str, optional): Deprecation message to be printed on the terminal.
            output_type (OutputTypes, optional): Override for the output type set in the config.
        """
        output_type_parsed = output_type or config.OUTPUT_TYPE
        table_params = config.TABLE_PARAMS

        if not config.DEBUG_MODE or self.ignore_debug_mode:
            if deprecation_message:
                self._print_deprecation_warning(deprecation_message)
            if output_type_parsed == OutputTypes.yaml.value:
                self.__yaml(data)
            elif output_type_parsed == OutputTypes.table.value:
                self.__table(data, table_params)
            else:
                self.__json(data, output_type)

    def __json(self, data: Any, output_type: OutputTypes | None = None) -> None:
        """Write logs to the terminal in JSON format.

        Args:
            data (Any): Data to be printed on the terminal.
            output_type (OutputTypes): If passed, JSON parser would be used to print output
                even if the data is not a dictionary.
        """
        if isinstance(data, (dict, list)) or output_type == OutputTypes.json:
            console.print_json(data=data)
        else:
            console.print(data)

    def __table(self, data: dict[Any, Any], params: dict[str, Any]) -> None:
        """Write logs to the terminal in a tabular format.

        Args:
            data (dict[Any, Any]): Data to be printed on the terminal.
            params (dict[str, Any]): Parameters for table-related formatting.
        """
        caption = (
            f"Page {data['pageNumber']} of {data['pageCount']}"
            if ("pageCount" in data and "pageNumber" in data)
            else None
        )

        if not len(data[params["data_key"]]):
            console.print(Panel(Text("No data to display.", justify="center")))
            return

        title = params["title"]

        if "subheader_key" in params and params["subheader_key"] in data:
            subheader_title = params.get("subheader_title", "Total")
            title = f'{title} ({subheader_title} = {data[params["subheader_key"]]})'

        table = Table(
            expand=True,
            highlight=True,
            row_styles=["grey62", ""],
            header_style="bold cyan",
            leading=1,
            title=title,
            title_style="bold",
            caption=caption,
        )

        output_keys = params["output_keys"].items()

        if len(output_keys) == 0:
            keys = data[params["data_key"]][0].keys()
            params["output_keys"] = {key: {"label": key} for key in keys}
            output_keys = params["output_keys"].items()

        for key_details in output_keys:
            table.add_column(key_details[1]["label"])

        for val in data[params["data_key"]]:
            parsed_values = []

            for key_details in output_keys:
                v = self.__get_data(val, key_details[0])

                if not v:
                    v = "-"
                elif "parser" in key_details[1]:
                    v = key_details[1]["parser"](v)

                if isinstance(v, (dict, list)):
                    parsed_value = json_highlighter(Text(json.dumps(v, indent=2), overflow="fold"))
                else:
                    parsed_value = Text(str(v), overflow="fold")

                parsed_values.append(parsed_value)

            table.add_row(*parsed_values)

        console.print("\n")
        console.print(table)
        console.print("\n")

    def __yaml(self, data: dict[Any, Any]) -> None:
        """Write logs to the terminal in YAML format.

        Args:
            data (dict[Any, Any]): Data to be printed on the terminal.
        """
        console.print(yaml.dump(data, indent=4))

    def __get_data(self, data: dict[Any, Any], key: str) -> Any:
        """Get value of a nested key from a dict.

        Args:
            data (dict[Any, Any]): The dictionary to get data from.
            key (str): The key to get the value for.

        Returns:
            Any: The value for the required key.
        """
        if not key:
            return data

        if key in data:
            return data[key]

        key_first_part = key.split(".")[0]
        key_rest = ".".join(key.split(".")[1:])

        if key_first_part not in data:
            return None

        return self.__get_data(data[key_first_part], key_rest)

    def pager(self) -> PagerContext | contextlib.nullcontext[None]:
        """Returns a Pager context manager."""
        if config.ENABLE_PAGING:
            return console.pager(styles=True)
        return contextlib.nullcontext()

    def _print_deprecation_warning(self, deprecation_message: str) -> None:
        """Prints a deprecation warning message."""
        print(f"\nNote: {deprecation_message}", file=sys.stderr)  # noqa: T201


__all__: List[str] = ["Writer"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#

"""Service client module."""
from __future__ import annotations

from typing import Any, Dict, Iterator, List, Literal, Optional, overload

from peak.base_client import BaseClient
from peak.constants import ContentType, HttpMethods
from peak.session import Session


class Service(BaseClient):
    """Client class for interacting with services resource."""

    BASE_ENDPOINT = "webapps/api/v1"

    @overload
    def list_services(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        status: Optional[List[str]] = None,
        name: Optional[str] = None,
        service_type: Optional[List[str]] = None,
        *,
        return_iterator: Literal[False],
    ) -> Dict[str, Any]: ...

    @overload
    def list_services(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        status: Optional[List[str]] = None,
        name: Optional[str] = None,
        service_type: Optional[List[str]] = None,
        *,
        return_iterator: Literal[True] = True,
    ) -> Iterator[Dict[str, Any]]: ...

    def list_services(
        self,
        page_size: Optional[int] = None,
        page_number: Optional[int] = None,
        status: Optional[List[str]] = None,
        name: Optional[str] = None,
        service_type: Optional[List[str]] = None,
        *,
        return_iterator: bool = True,
    ) -> Iterator[Dict[str, Any]] | Dict[str, Any]:
        """Retrieve a list of services.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/webapps/api-docs/index.htm#/Services/list-service>`__

        Args:
            page_size (int | None): The number of services per page.
            page_number (int | None): The page number to retrieve. Only used when `return_iterator` is False.
            status (List[str] | None): A list of service status to filter the list by.
                Valid values are `CREATING`, `DEPLOYING`, `AVAILABLE`, `DELETING`, `CREATE_FAILED`, `DELETE_FAILED`.
            name (str | None): Name of the service to search for.
            service_type (List[str] | None): A list of service types to filter the list by. Valid values are `api`, `web-app` and `shiny`.
            return_iterator (bool): Whether to return an iterator object or list of services for a specified page number, defaults to True.

        Returns:
            Iterator[Dict[str, Any] | Dict[str, Any]: an iterator object which returns an element per iteration, until there are no more elements to return.
            If `return_iterator` is set to False, a dictionary containing the list and pagination details is returned instead.
            Set `return_iterator` to True if you want automatic client-side pagination, or False if you want server-side pagination.

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/webapps/"
        params = {
            "pageSize": page_size,
            "status": status,
            "searchTerm": name,
            "serviceType": service_type,
            "featureType": "services",
        }

        if return_iterator:
            return self.session.create_generator_request(
                endpoint,
                method,
                content_type=ContentType.APPLICATION_JSON,
                response_key="services",
                params=params,
            )

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            params={**params, "pageNumber": page_number},
        )

    def create_service(self, body: Dict[str, Any]) -> Dict[str, str]:
        """Create a new service.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/webapps/api-docs/index.htm#/Services/create-service>`__

        Args:
            body (Dict[str, Any]): A dictionary containing the service config. Schema can be found below.

        Returns:
            Dict[str, str]: Id of the new service

        SCHEMA:
            .. code-block:: json

                {
                    "name": "string(required)",
                    "title": "string",
                    "serviceType": "string. Valid values are 'api', 'web-app' and 'shiny'",
                    "imageDetails": {
                        "imageId": "number(required)",
                        "versionId": "number"
                    },
                    "resources": {
                        "instanceTypeId": "number"
                    },
                    "parameters": {
                        "env": {
                            "key: string": "value: string"
                        },
                        "secrets": []
                    },
                    "description": "string",
                    "sessionStickiness": "boolean. Not required for 'api' service type.",
                    "scaleToZero": "boolean. Only for 'web-app' service type.",
                    "entrypoint": "string",
                    "healthCheckURL": "string",
                    "minInstances": "number. Default is 1 and maximum is 2",
                }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/webapps/"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
        )

    def update_service(
        self,
        service_id: str,
        body: Dict[str, Any],
    ) -> Dict[str, str]:
        """Updates the existing service.

        When updating the service, it will trigger a redeployment only under specific conditions.
        Redeployment is triggered if you make changes to any of the following parameters: imageId, versionId, instanceTypeId, parameters, healthCheckURL, entrypoint, scaleToZero or sessionStickiness.
        However, only modifying the title or description will not trigger a redeployment.

        With the help of this operation, we can just update the required fields (except name and serviceType) and keep the rest of the fields as it is.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/webapps/api-docs/index.htm#/Services/update-service>`__

        Args:
            service_id (str): The ID of the service to update.
            body (Dict[str, Any]): A dictionary containing the service config. Schema can be found below.

        Returns:
            Dict[str, str]: Id of the service.

        SCHEMA:
            .. code-block:: json

                {
                    "title": "string",
                    "imageDetails": {
                        "imageId": "number(required)",
                        "versionId": "number",
                    },
                    "resources": {
                        "instanceTypeId": "number"
                    },
                    "parameters": {
                        "env": {
                            "key: string": "value: string",
                        },
                        "secrets": []
                    },
                    "description": "string",
                    "sessionStickiness": "boolean. Not required for 'api' service type.",
                    "scaleToZero": "boolean. Only for 'web-app' service type.",
                    "entrypoint": "string",
                    "healthCheckURL": "string",
                    "minInstances": "number. Default is 1 and maximum is 2",
                }

        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given service does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.PATCH, f"{self.BASE_ENDPOINT}/webapps/{service_id}"
        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
        )

    def create_or_update_service(self, body: Dict[str, Any]) -> Dict[str, str]:
        """Create a new service or updates an existing service based on service name.

        When updating the service, it will trigger a redeployment only under specific conditions.
        Redeployment is triggered if you make changes to any of the following parameters: imageId, versionId, instanceTypeId, parameters, healthCheckURL, entrypoint, scaleToZero or sessionStickiness.
        However, only modifying the title or description will not trigger a redeployment.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/webapps/api-docs/index.htm#/Services/create-service>`__

        Args:
            body (Dict[str, Any]): A dictionary containing the service config. Schema can be found below.

        Returns:
            Dict[str, str]: Id of the new or updated service.

        SCHEMA:
            .. code-block:: json

                {
                    "name": "string(required)",
                    "title": "string",
                    "serviceType": "string. Valid values are 'api', 'web-app' and 'shiny'",
                    "imageDetails": {
                        "imageId": "number(required)",
                        "versionId": "number"
                    },
                    "resources": {
                        "instanceTypeId": "number"
                    },
                    "parameters": {
                        "env": {
                            "key: string": "value: string",
                        },
                        "secrets": []
                    },
                    "description": "string",
                    "sessionStickiness": "boolean. Not required for 'api' service type.",
                    "scaleToZero": "boolean. Only for 'web-app' service type.",
                    "entrypoint": "string",
                    "healthCheckURL": "string",
                    "minInstances": "number. Default is 1 and maximum is 2",
                }


        Raises:
            BadRequestException: The given request parameters are invalid.
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            InternalServerErrorException: The server failed to process the request.
        """
        service_name = body.get("name", "")
        response = (
            {} if not len(service_name) else self.list_services(page_size=100, return_iterator=False, name=service_name)
        )
        filtered_services = list(
            filter(lambda service: service.get("name", "") == service_name, response.get("services", [])),
        )

        if len(filtered_services) > 0:
            service_id = filtered_services[0]["id"]
            return self.update_service(service_id=service_id, body=body)

        return self.create_service(body=body)

    def delete_service(
        self,
        service_id: str,
    ) -> Dict[str, str]:
        """Delete a service.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/webapps/api-docs/index.htm#/Services/delete-service>`__

        Args:
            service_id (str): The ID of the service to delete.

        Returns:
            Dict[str, str]: Dictonary containing ID of the deleted service.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given service does not exist.
            ConflictException: If the service is in a conflicting state while deleting.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.DELETE, f"{self.BASE_ENDPOINT}/webapps/{service_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def describe_service(
        self,
        service_id: str,
    ) -> Dict[str, Any]:
        """Retrieve details of a specific service.

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/webapps/api-docs/index.htm#/Services/get-service>`__

        Args:
            service_id (str): The ID of the service to fetch.

        Returns:
            Dict[str, Any]: Dictonary containing details of the service.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given service does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.GET, f"{self.BASE_ENDPOINT}/webapps/{service_id}"

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
        )

    def test_service(
        self,
        service_name: str,
        http_method: str,
        path: Optional[str] = None,
        payload: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Test an API service to verify it's health and if it is working. Make sure the service is of type "api".

        REFERENCE:
            🔗 `API Documentation <https://service.peak.ai/webapps/api-docs/index.htm#/Services/test-api-service>`__

        Args:
            service_name (str): The name of the API type service to test.
            http_method (str): The HTTP method to use for the test. Valid values are 'get', 'post', 'put', 'patch' and 'delete'.
            path (str, None): The path to test the service with. Default is '/'.
            payload (Dict[str, Any], None): The payload to test the service with.

        Returns:
            Dict[str, Any]: Dictonary containing response of the test.

        Raises:
            UnauthorizedException: The credentials are invalid.
            ForbiddenException: The user does not have permission to perform the operation.
            NotFoundException: The given service does not exist.
            UnprocessableEntityException: The server was unable to process the request.
            InternalServerErrorException: The server failed to process the request.
        """
        method, endpoint = HttpMethods.POST, f"{self.BASE_ENDPOINT}/webapps/{service_name}/test"
        body = {"httpMethod": http_method, "path": path, "payload": payload}

        return self.session.create_request(  # type: ignore[no-any-return]
            endpoint,
            method,
            content_type=ContentType.APPLICATION_JSON,
            body=body,
        )


def get_client(session: Optional[Session] = None) -> Service:
    """Returns a Service client, If no session is provided, a default session is used.

    Args:
        session (Optional[Session]): A Session Object. Default is None.

    Returns:
        Service: The service client object.
    """
    return Service(session)


__all__ = ["get_client"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Logger for the Peak SDK."""
from __future__ import annotations

import logging
import sys
from typing import List

from peak.constants import LOG_FORMAT, LOG_LEVELS

logger: logging.Logger = logging.getLogger("peak-sdk")
stream_handler = logging.StreamHandler(sys.stdout)
stream_handler.setFormatter(LOG_FORMAT)
logger.addHandler(stream_handler)
logger.propagate = False


def set_log_level(log_level: LOG_LEVELS) -> None:
    """Update log level for Peak logger.

    Args:
        log_level (LOG_LEVELS): new logging level for the logger.
    """
    logger.setLevel(log_level)
    for handler in logger.handlers:
        handler.setLevel(log_level)


__all__: List[str] = ["logger"]



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""CLI module for Peak Metrics."""



#
# # Copyright © 2024 Peak AI Limited. or its affiliates. All Rights Reserved.
# #
# # Licensed under the Apache License, Version 2.0 (the "License"). You
# # may not use this file except in compliance with the License. A copy of
# # the License is located at:
# #
# # https://github.com/PeakBI/peak-sdk/blob/main/LICENSE
# #
# # or in the "license" file accompanying this file. This file is
# # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# # ANY KIND, either express or implied. See the License for the specific
# # language governing permissions and limitations under the License.
# #
# # This file is part of the peak-sdk.
# # see (https://github.com/PeakBI/peak-sdk)
# #
# # You should have received a copy of the APACHE LICENSE, VERSION 2.0
# # along with this program. If not, see <https://apache.org/licenses/LICENSE-2.0>
#
"""Peak blocks deployments commands."""

from __future__ import annotations

from typing import Dict, List, Optional

import typer
from peak.cli import args, helpers
from peak.cli.args import DRY_RUN, GENERATE_YAML, OUTPUT_TYPES, PAGING
from peak.constants import OutputTypes, OutputTypesNoTable
from peak.output import Writer
from peak.press.blocks import Block

app = typer.Typer()

_DEPLOYMENT_ID = typer.Argument(..., help="ID of the Block deployment to be used in this operation")

_DEPLOYMENT_ID_OPTION = typer.Option(None, help="ID of the Block deployment to be used in this operation")
_FALLBACK_PARAMS_FILE = typer.Option(
    None,
    help="File containing parameters to be used when deployment id is not present",
)
_FALLBACK_DETAILS_FILE = typer.Option(
    None,
    help="File containing details to be used when deployment id is not present",
)


@app.command(short_help="Create a Block deployment.")
def create(
    ctx: typer.Context,
    file: str = args.TEMPLATE_PATH,
    description_file: Optional[str] = args.TEMPLATE_DESCRIPTION_FILE,
    revision_notes_file: Optional[str] = args.REVISION_NOTES_FILE,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a Block deployment This creates the resource described in the Block Spec.

    \b
    🧩 ***Input file schema (yaml):***<br/>
    ```yaml
      body (map):
        metadata (map):
            name (string | required: false): Name of the deployment. Must be unique within the tenant.
            title (string | required: false): Title of the deployment.
            summary (string | required: false): Summary of the deployment.
            description (string | required: false): Description of the deployment.
            descriptionContentType (string | required: false): Content type of the description. Should be one of "text/plain" or "text/markdown".
            imageUrl (string | required: false): URL of the image to be associated with the block deployment.
            tags (list(map) | required: false):
                - name (string): Name of the tag.
        parameters (map | required: false):
            <parameterType> (map): Dictionary of parameters. Here the key is the type of the parameter. Accepted values are "build" or "run".
                <parameterName> (string | boolean | number | list(string)): Value of the parameter. Here the key is the name of the parameter.
        revision (map | required: false):
            notes (string | required: false): Notes for the deployment revision.
        spec (map):
            id (string): ID of the block spec to be deployed.
            release (map | required: false):
                version (string): A valid semantic release version of the block spec.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak blocks deployments create /path/to/body.yaml -v /path/to/params.yaml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb"
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/post_v1_blocks_deployments)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    markdown_files = {}
    if description_file:
        markdown_files["body:metadata:description"] = description_file
    if revision_notes_file:
        markdown_files["body:revision:notes"] = revision_notes_file

    body = helpers.template_handler(file, params_file, params, markdown_files)
    body = helpers.remove_unknown_args(body, blocks_client.create_deployment)

    with writer.pager():
        response: Dict[str, str] = blocks_client.create_deployment(**body)
        writer.write(response)


@app.command("list", short_help="List Block deployments.")
def list_block_deployments(
    ctx: typer.Context,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    status: Optional[List[str]] = args.STATUS_FILTER_DEPLOYMENTS,
    name: Optional[str] = args.NAME_FILTER,
    title: Optional[str] = args.TITLE_FILTER,
    sort: Optional[List[str]] = args.SORT_KEYS,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** Block deployments that have been created for the given tenant.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments list --status=deployed,failed --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "deploymentsCount": 1,
        "deployments": [...],
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 25
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/get_v1_blocks_deployments)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.list_deployments(
            status=status,
            name=name,
            sort=sort,
            title=title,
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)


@app.command(short_help="Describe the Block deployment.")
def describe(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** the Block deployment with details of its latest revision.

    \b
    ***Note***: Parameters listed in the response are masked if "hideValue" was set to true
        when creating the block spec.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments describe <deployment_id>
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb",
        "kind": "app",
        "latestRevision": {...},
        "metadata": {...},
        "spec": {...}
        "parameters": {...}
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/get_v1_blocks_deployments__deploymentId_)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.describe_deployment(deployment_id)
        writer.write(response)


@app.command(short_help="Update the Block deployment metadata.")
def update_metadata(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    file: str = args.TEMPLATE_PATH,
    description_file: Optional[str] = args.TEMPLATE_DESCRIPTION_FILE,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Update*** the Block deployment metadata.

    \b
    🧩 ***Input file schema (yaml):***<br/>
    ```yaml
      body (map):
        name (string | required: false): Name of the deployment. Must be unique within the tenant.
        title (string | required: false): Title of the deployment.
        summary (string | required: false): Summary of the deployment.
        description (string | required: false): Description of the deployment.
        descriptionContentType (string | required: false): Content type of the description. Should be one of "text/plain" or "text/markdown".
        imageUrl (string | required: false): URL of the image to be associated with the block deployment.
        tags (list(map) | required: false):
            - name (string): Name of the tag.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak blocks deployments update-metadata <deployment_id> /path/to/body.yaml -v /path/to/params.yaml
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb"
        "kind": "workflow",
        "latestRevision": {...},
        "metadata": {...},
        "spec": {...}
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/patch_v1_blocks_deployments__deploymentId_)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    markdown_files = {}
    if description_file:
        markdown_files["body:description"] = description_file

    body = helpers.template_handler(file, params_file, params, markdown_files)
    body = helpers.remove_unknown_args(body, blocks_client.update_deployment_metadata)

    with writer.pager():
        response = blocks_client.update_deployment_metadata(deployment_id, **body)
        writer.write(response)


@app.command(short_help="Redeploy a Block deployment.")
def redeploy(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Redeploy*** latest revision of a Block deployment. This allows you to redeploy a Block deployment that is in `failed` or `warning` state.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments redeploy <deployment_id>
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "deploymentId": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb",
        "revision": 2
        "revisionId": "7092bd84-c35d-43c1-90ca-7510a1204dcc"
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/redeploy_v1_blocks_deployments__deploymentId_)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.redeploy(deployment_id)
        writer.write(response)


@app.command(short_help="Delete a Block deployment.")
def delete(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Delete*** a Block deployment. This will delete the resource that was created as part of the deployment.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments delete <deployment_id>
    ```

    \b
    🆗 ***Response:***
    ```json
    {}
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/delete_v1_blocks_deployments__deploymentId_)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.delete_deployment(deployment_id)
        writer.write(response)


@app.command(short_help="Create a Block deployment revision.")
def create_revision(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    file: str = args.TEMPLATE_PATH,
    revision_notes_file: Optional[str] = args.REVISION_NOTES_FILE,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Create*** a Block deployment revision. This updates the deployment with the specified spec release.

    \b
    🧩 ***Input file schema (yaml):***<br/>
    ```yaml
      body (map):
        parameters (map | required: false):
            <parameterType> (map): Dictionary of parameters. Here the key is the type of the parameter. Accepted values are "build" or "run".
                <parameterName> (string | boolean | number | list(string)): Value of the parameter. Here the key is the name of the parameter.
        release (map):
            version (string): A valid semantic release version of the block spec.
        revision (map | required: false):
            notes (string | required: false): Notes for the deployment revision.

    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak blocks deployments create-revision /path/to/body.yaml -v /path/to/params.yaml
    ```

    \b
    🆗 ***Response:***
    ```json
    {
        "id": "632a4e7c-ab86-4ecb-8f34-99b5da531ceb",
        "revision": 2
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/post_v1_blocks_deployments__deploymentId__revisions)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    markdown_files = {}
    if revision_notes_file:
        markdown_files["body:revision:notes"] = revision_notes_file

    body = helpers.template_handler(file, params_file, params, markdown_files)
    body = helpers.remove_unknown_args(body, blocks_client.create_deployment_revision)

    with writer.pager():
        response: Dict[str, str] = blocks_client.create_deployment_revision(deployment_id, **body)
        writer.write(response)


@app.command(short_help="Describe a Block deployment revision.")
def describe_revision(
    ctx: typer.Context,
    deployment_id: str = args.DEPLOYMENT_ID,
    revision: str = args.REVISION,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Describe*** a Block deployment revision.

    \b
    ***Note***: Parameters listed in the response are masked if "hideValue" was set to true
        when creating the block spec release.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments describe-revision --deployment-id <deployment_id> --revision 2
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "createdAt": "2020-01-01T18:00:00.000Z",
        "createdBy": "jane.smith@peak.ai",
        "id": "7092bd84-c35d-43c1-90ca-7510a1204dcc",
        "latestRevision": {...},
        "notes": "This is a new revision"
        "resources": [...],
        "revision": 2,
        "status": "deploying",
        "spec": {...}
        "parameters": {...}
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/get_v1_blocks_deployments__deploymentId__revisions__revision_)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.describe_deployment_revision(deployment_id, revision)
        writer.write(response)


@app.command(short_help="List revisions of a Block deployment.")
def list_revisions(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    sort: Optional[List[str]] = args.SORT_KEYS,
    status: Optional[List[str]] = args.STATUS_FILTER_DEPLOYMENT_REVISIONS,
    page_size: Optional[int] = args.PAGE_SIZE,
    page_number: Optional[int] = args.PAGE_NUMBER,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypes] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***List*** all revisions for a given Block deployment.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments list-revisions "632a4e7c-ab86-4ecb-8f34-99b5da531ceb" --sort createdBy:asc,createdAt --status deployed,deploying --page-size 10 --page-number 1
    ```

    \b
    🆗 ***Response:***
    ```
    {
        "pageCount": 1,
        "pageNumber": 1,
        "pageSize": 10,
        "revisionCount": 1,
        "revisions": [...]
    }
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Block%20Deployments/get_v1_blocks_deployments__deploymentId__revisions)
    """
    block_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = block_client.list_deployment_revisions(
            deployment_id,
            sort=sort,
            status=status,
            page_size=page_size,
            page_number=page_number,
            return_iterator=False,
        )
        writer.write(response)


@app.command(short_help="Get the parameters for a deployment at run time.")
def get_parameters(
    ctx: typer.Context,
    deployment_id: Optional[str] = _DEPLOYMENT_ID_OPTION,
    fallback_params_file: Optional[str] = _FALLBACK_PARAMS_FILE,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Get*** all the parameters for a deployment at run time.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments get-parameters --deployment-id=<deployment-id> --fallback-params-file=<path/to/fallback/params/file>
    ```

    \b
    🆗 ***Response:***
    ```
    {...}
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Deployment%20Parameters/get_v1_deployments__deploymentId__parameters_run)
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.get_parameters(
            deployment_id=deployment_id,
            fallback_params_file=fallback_params_file,
        )
        writer.write(response)


@app.command(short_help="Update the parameters for a deployment at run time.")
def patch_parameters(
    ctx: typer.Context,
    deployment_id: str = _DEPLOYMENT_ID,
    file: str = args.TEMPLATE_PATH,
    params_file: Optional[str] = args.TEMPLATE_PARAMS_FILE,
    params: Optional[List[str]] = args.TEMPLATE_PARAMS,
    dry_run: Optional[bool] = DRY_RUN,  # noqa: ARG001
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
    generate: Optional[bool] = GENERATE_YAML,  # noqa: ARG001
) -> None:
    """***Update*** the parameters for a deployment at run time.

    \b
    🧩 ***Input file schema (yaml):***<br/>
    ```yaml
    body (map):
        <key> (string | boolean | number | list(string)): Value of the parameter. Here the key is the name of the parameter.
    ```

    \b
    📝 ***Example usage:***
    ```bash
    peak blocks deployments patch-parameters <deployment-id> /path/to/body.yaml -v /path/to/params.yaml
    ```

    \b
    🆗 ***Response:***
    ```
    {...}
    ```

    🔗 [**API Documentation**](https://press.peak.ai/api-docs/index.htm#/Deployment%20Parameters/patch_v1_deployments__deploymentId__parameters_run)
    """
    body = helpers.template_handler(file, params_file, params)
    blocks_client: Block = ctx.obj["client"]
    body = helpers.remove_unknown_args(body, blocks_client.patch_parameters)
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.patch_parameters(deployment_id, **body)
        writer.write(response)


@app.command(short_help="Get the info for related blocks within an app.")
def get_related_block_details(
    ctx: typer.Context,
    deployment_id: Optional[str] = _DEPLOYMENT_ID_OPTION,
    fallback_details_file: Optional[str] = _FALLBACK_DETAILS_FILE,
    paging: Optional[bool] = PAGING,  # noqa: ARG001
    output_type: Optional[OutputTypesNoTable] = OUTPUT_TYPES,  # noqa: ARG001
) -> None:
    """***Get*** the info for related blocks within an app.

    \b
    📝 ***Example usage:***<br/>
    ```bash
    peak blocks deployments get_related_block_details --deployment-id=<deployment-id> --fallback-details-file=<path/to/fallback/details/file>
    ```

    \b
    🆗 ***Response:***
    ```
    {...}
    ```
    """
    blocks_client: Block = ctx.obj["client"]
    writer: Writer = ctx.obj["writer"]

    with writer.pager():
        response = blocks_client.get_related_block_details(
            deployment_id=deployment_id,
            fallback_details_file=fallback_details_file,
        )
        writer.write(response)
